function [LL LH dinfo pinfo] = QDLifting_2(L, lsi)
% Stage 2 decomposition: decompose LL band into LL band and 
% LH band with directional lifting on qincux lattice
%
% USAGE:
%    [LL LH dinfo pinfo] = QDLifting_2(L, lsi)
% INPUT:   
%     L   - Lower band in stage 1 decomposition
%           
%    lsi  - Lifting lattice
% OUTPUT:
%    LL   - Lower-Lower band
%    HH   - Lower-Higher band
%  dinfo  - directional information 
%  pinfo  - partition information
% 
% REFERENCE:
%    Chuo-Ling Chang, Arian Maleki, and Bernd Girod, ˇ°Adaptive
%    wavelet transform for image for image compression via
%    directional quincux lifting", in Proc. IEEE International 
%    Workshop on Multimedia Signal Processing, MMSP-05, Shanghai,
%    China, November 2005.

% Lazy wavelet 
y = QSplit_2(L);  

N = size(lsi, 1);
for i = 1:(N-1)
     f  = lsi{i,2}; % filter taps
     fd = lsi{i,3}; % filter delay: z(-fd)
     cd = lsi{i,4}; % component delay which corresponds non-one diagonal components: z(-cd)
    if lsi{i,1}=='p'
        [y{2}, dinfo, pinfo] = Predict_2(y, f, fd);
    elseif lsi{i,1}=='u'
         y{1} = Update(y, f, fd, dinfo, pinfo);
    end
end
y = Normalize(y, lsi{N,1}, lsi{N,2});
LL = y{1};
LH = y{2};


function y = Normalize(x, Lscale, Hscale)
y{1} = Lscale*x{1};
y{2} = Hscale*x{2};

function y  =  QSplit_2(L)
% Qincux sampling
%   Output:      y - Result of lazy wavelet
%             y{1} - Even field of L
%             y{2} - Odd field of L
%   Notes: I don't know why this paper take the odd 
%          samples (dashed dots) as even field and take 
%          even samples (solid dots) as odd field
% 

y{1} = L(:, 2:2:end);
y{2} = L(:, 1:2:end);

function [h dinfo pinfo] = Predict_2(x, f, fd)
% primal lift: predicting along conlumns
%  direction indices and corresponding degrees
%   didx:      1    2    3    4    5    6    7     8
%   degree:    0  -27  -45  -63  -90 -117  -135  -153 
s = x{1};
d = x{2};
[r c]= size(d);
dset1 = [0 0; 0   1; 0  0; 1  0; 0 0;   1 -1; 0 -1; 0 -2];
dset2 = [0 0; -1 -2; -1 -1;-2 -1; 0 0; -2  0; -1 0; -1 1];
ldset = length(dset1);

h_all = zeros(r, c, ldset);
h = zeros(r, c);
bw = 1;
bh = 2;
br = r/bh;
bc = c/bw;
hsad = zeros(br, bc, ldset);
minsad = zeros(br,bc);
dinfo = zeros(br, bc);
pinfo = zeros(br, bc);
rextn = 2;
cextn = 2;

% symmetric extension
extns  = extn2D(s',1:(rextn+1), 'symodd')'; % odd symmtric extension with rextn samples for each row
extns  = extn2D(extns, 1:(cextn+1), 'symodd'); % odd symmtric extension with cextn samples for each row 
% generate 
% Predict for all spatial directions except directiona 
% 1 and 5 since they lie in half pixel position

for i = 1:ldset;
%     if (i~=1&i~=5)
    h_all(1:r, :, i) = d(1:r, :) - 0.5*extns( (1:r)+cextn + dset1(i,1), (1:c)+rextn + dset1(i,2) ) ...
                                 - 0.5*extns( (1:r)+cextn + dset2(i,1), (1:c)+rextn + dset2(i,2) );
%     end
end

% generate half pixel value for direction 1
halfh1 = zeros(r, 1+c);
i = 1:r;
j = 1:(1+c);
        half1(i, j) = (-extns(cextn + i - 2, rextn + j -1) + 3*extns(cextn + i - 1, rextn + j -1) + ...
            3*extns(cextn + i , rextn + j - 1) - extns(cextn + i + 1, rextn + j - 1) )/4;
% predict for direction 1
    h_all(1:r, 1:c, 1) = d(1:r, 1:c) - 0.5*half1(1:r, 1:c) - 0.5*half1(1:r, (1:c)+1);
    
% generate half pixel value for direction 5
halfh5 = zeros(1+r, c);
i = 1:(1+r);
j = 1:c;
        half5(i, j) = (-extns(cextn + i - 1, rextn + j - 2) + 3*extns(cextn + i -1, rextn + j -1) + ...
            3*extns(cextn + i - 1, rextn + j ) - extns(cextn + i - 1, rextn + j + 1) )/4;

% predict for direction 5
    h_all(1:r, 1:c, 5) = d(1:r, 1:c) - 0.5*half5(1:r, 1:c) - 0.5*half5((1:r)+1, (1:c));

% finding the best direction which results in 
% smallest residue in SAD
for i = 1:br
    for j = 1:bc
        roff = (i-1)*bh;
        coff = (j-1)*bw;
        hsad(i, j, 1:ldset) = sum(sum(abs(h_all((roff+1):(roff+bh), (coff+1):(coff+bw), 1:ldset)),1),2);
%         hsad(i, j, 1:ldset) = 10*round(hsad(i, j, 1:ldset)/10); %quantized object function
        [minsad(i,j) dinfo(i,j)] = min(hsad(i, j, :));
        h((roff+1):(roff+bh), (coff+1):(coff+bw)) = h_all((roff+1):(roff+bh), (coff+1):(coff+bw), dinfo(i,j));
    end
end
%  [minsad, dinfo] = min(hsad,[], 3);
 

function [l] = Update(x, f, fd, dinfo, pinfo)
% dual lift: updating along conlumns
s = x{1};
d = x{2};
[r c]= size(d);
dset1 = [0 0;  1  2; 1  1; 2  1; 0 0;  2  0; 1 0; 1 -1];
dset2 = [0 0;  0 -1; 0  0;-1  0; 0 0; -1  1; 0 1; 0  2];
ldset = length(dset1);
l_all = zeros(r, c, ldset);
l = zeros(r, c);
bw = 1;
bh = 2;
br = r/bh;
bc = c/bw;
rextn = 2;
cextn = 2;

% symmetric extension
extnd  = extn2D(d',1:(rextn+1), 'symodd')'; % odd symmtric extension with rextn samples for each row
extnd  = extn2D(extnd, 1:(cextn+1), 'symodd'); % odd symmtric extension with cextn samples for each row 

% update for all spatial directions except directiona 
% 1 and 5 since they lie in half pixel position
for i = 1:ldset;
%     if (i~=1&i~=5)
    l_all(1:r, :, i) = s(1:r, :) + 0.25*extnd( (1:r)+cextn + dset1(i,1), (1:c)+rextn + dset1(i,2) ) ...
                                 + 0.25*extnd( (1:r)+cextn + dset2(i,1), (1:c)+rextn + dset2(i,2) );
%     end
end
% generate half pixel value for direction 1
halfh1 = zeros(r, 1+c);
i = 1:r;
j = 1:(1+c);
        half1(i, j) = (-extnd(cextn + i - 1, rextn + j) + 3*extnd(cextn + i, rextn + j) + ...
            3*extnd(cextn + i + 1, rextn + j) - extnd(cextn + i + 2, rextn + j) )/4;
% update for direction 1
    l_all(1:r, 1:c, 1) = s(1:r, 1:c) + 0.25*half1(1:r, 1:c) + 0.25*half1(1:r, (1:c)+1);
    
% update half pixel value for direction 5
halfh5 = zeros(1+r, c);
i = 1:(1+r);
j = 1:c;
        half5(i, j) = (-extnd(cextn + i, rextn + j - 1) + 3*extnd(cextn + i, rextn + j) + ...
            3*extnd(cextn + i, rextn + j + 1) - extnd(cextn + i, rextn + j + 2) )/4;
% update for direction 5
    l_all(1:r, 1:c, 5) = s(1:r, 1:c) + 0.25*half5(1:r, 1:c) + 0.25*half5((1:r)+1, (1:c));
  

% Pick the right one with dinfo
for i = 1:br
    for j = 1:bc
        roff = (i-1)*bh;
        coff = (j-1)*bw;
        l((roff+1):(roff+bh), (coff+1):(coff+bw)) = l_all((roff+1):(roff+bh), (coff+1):(coff+bw), dinfo(i,j));
    end
end
